
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "myhtml.h"

enum {START, TAG, CHAR, END};

static hnode *h, *ch;

static char *b = NULL;

static int n, nmax = 0;

static struct {
        char *name;
        int value;
} cchar[] = {
        {"quot", '"'}, {"amp", '&'}, {"lt", '<'}, {"gt", '>'},
        {"nbsp", 160}, {"iexcl", 161}, {"cent", 162}, {"pound", 163},
        {"curren", 164}, {"yen", 165}, {"brvbar", 166}, {"sect", 167},
        {"uml", 168}, {"copy", 169}, {"ordf", 170}, {"laquo", 171},
        {"not", 172}, {"shy", 173}, {"reg", 174}, {"macr", 175},
        {"deg", 176}, {"plusmn", 177}, {"sup2", 178}, {"sup3", 179},
        {"acute", 180}, {"micro", 181}, {"para", 182}, {"middot", 183},
        {"cedil", 184}, {"sup1", 185}, {"ordm", 186}, {"raquo", 187},
        {"frac14", 188}, {"frac12", 189}, {"frac34", 190}, {"iquest", 191},
        {"Agrave", 192}, {"Aacute", 193}, {"Acirc", 194}, {"Atilde", 195},
        {"Auml", 196}, {"Aring", 197}, {"AElig", 198}, {"Ccedil", 199},
        {"Egrave", 200}, {"Eacute", 201}, {"Ecirc", 202}, {"Euml", 203},
        {"Igrave", 204}, {"Iacute", 205}, {"Icirc", 206}, {"Euml", 207},
        {"ETH", 208}, {"Ntilde", 209}, {"Ograve", 210}, {"Oacute", 211},
        {"Ocirc", 212}, {"Otilde", 213}, {"Ouml", 214}, {"times", 215},
        {"Oslash", 216}, {"Ugrave", 217}, {"Uacute", 218}, {"Ucirc", 219},
        {"Uuml", 220}, {"Yacute", 221}, {"THORN", 222}, {"szlig", 223},
        {"agrave", 224}, {"aacute", 225}, {"acirc", 226}, {"atilde", 227},
        {"auml", 228}, {"aring", 229}, {"aelig", 230}, {"ccedil", 231},
        {"egrave", 232}, {"eacute", 233}, {"ecirc", 234}, {"euml", 235},
        {"igrave", 236}, {"iacute", 237}, {"icirc", 238}, {"iuml", 239},
        {"eth", 240}, {"ntilde", 241}, {"ograve", 242}, {"oacute", 243},
        {"ocirc", 244}, {"otilde", 245}, {"ouml", 246}, {"divide", 247},
        {"slash", 248}, {"ugrave", 249}, {"uacute", 250}, {"ucirc", 251},
        {"uuml", 252}, {"yacute", 253}, {"thorn", 254}, {"yuml", 255},
        {NULL, 0}
};

/* convert Auml to  and #33 to ! */
static int from_cchar(char *from)
{
        int i;
 
        if (from[0] == '#') {
                i = atoi(from+1);
                if (i >= ' ' && i <= 255) return i;
                return -1;
        }
        for (i = 0; cchar[i].name; i++) {
                if (!strcmp(cchar[i].name, from)) return cchar[i].value;
        }
        return -1;
}

static hnode *new_hnode(void)
{
	hnode *h = malloc(sizeof *h);
	h->tag = NULL;
	h->text = NULL;
	h->parent = NULL;
	h->next = NULL;
	h->child = NULL;
	return h;
}

static void indent(int i)
{
	int j;

	for (j = 0; j < i; j++) putchar(' ');
}

void free_html(hnode *h)
{
	if (h == NULL) return;
	if (h->tag) free(h->tag);
	if (h->text) free(h->text);
	free_html(h->next);
	free_html(h->child);
}

void dump_html(hnode *h, int i)
{
	if (h == NULL) return;
	if (h->tag == NULL) {
		indent(i);
		printf("Text: '%s'\n", h->text);
	} else {
		indent(i);
		printf("Tag: '%s'\n", h->tag);
		if (h->text) {
			indent(i+4);
			printf("Value: '%s'\n", h->text);
		}
	}
	dump_html(h->child, i+2);
	dump_html(h->next, i);
}

static void growb(int n)
{
	if (n >= nmax) {
		nmax = n+1000;
		b = realloc(b, nmax+1);
	}
}

static void clearb(void)
{
	n = 0;
	growb(n+1);
	b[n] = '\0';
}

static hnode *last_node(hnode *h)
{
	if (h == NULL) return NULL;
	if (h->next == NULL) return h;
	return last_node(h->next);
}

/*
When we enter ins_node, ch points to the most recently inserted node.
The new node is inserted as a child to this node. When we are done,
ch points to the newly inserted node.

Special case: If tree is empty, ch is NULL. In that case, the new
node becomes the root of the tree.
*/

static void ins_node(hnode *n)
{
	hnode *ph;

	if (ch == NULL) {
		h = n;
	} else if ((ph = last_node(ch->child)) == NULL) {
		ch->child = n;
	} else {
		ph->next = n;
	}
	n->parent = ch;
	ch = n;
}

static hnode *match_tag(char *p)
{
	hnode *h;

	for (h = ch; h; h = h->parent) {
		if (h->tag && !strcmp(h->tag, p)) break;
	}
	return h;
}

static void trim(char *p)
{
	char *q = p;

	if (p == NULL) return;
	while (isspace(*q)) q++;
	memcpy(p, q, strlen(q)+1);
}

static void tag_cb(char *p)
{
	hnode *h;
	char *q;
	char b[1024];
	int n;

	if (p == NULL) return;		/* should never happen */
	q = p;
	while (isspace(*q)) q++;
	if (p[0] == '\0') return;	/* ignore empty tag */
	n = 0;
	while (*q && !isspace(*q)) b[n++] = toupper(*q++);
	b[n] = '\0';
	while (*q && isspace(*q)) q++;	/* skip past white space */
	if (p == NULL || p[0] == '\0') return;
	if (p[0] == '/') {
		/* closing tag */
		h = match_tag(b+1);
		if (h == NULL) {
			fprintf(stderr, "Warning: bogus closer '%s'\n", p);
		} else {
			ch = h->parent;
		}
	} else {
		h = new_hnode();
		h->tag = strdup(b);
		if (*q) h->text = strdup(q);
		ins_node(h);
	}
}

static int extchar(char *p)
{
	int c = from_cchar(p);
	if (c == -1) c = '?';
	return c;
}

static void text_cb(char *p)
{
	hnode *h;

	trim(p);
	if (p == NULL || p[0] == '\0') return;
	h = new_hnode();
	h->text = strdup(p);
	ins_node(h);
	ch = ch->parent;
}

static void char_cb(int c)
{
	growb(n+1);
	b[n++] = c;
	b[n] = '\0';
}

hnode *parse_html(int (*getc_cb)(void *), void *closure)
{
	int state = 0, error = 0;
	int c;
	char t[1024];
	int m;

	ch = h = NULL;
	n = 0;
	while (state != END) {
		c = getc_cb(closure);
		switch (state) {
		case START:
			if (c == '<') {
				text_cb(b);
				clearb();
				state = TAG;
			} else if (c == '&') {
				m = 0;
				state = CHAR;
			} else if (c == '\0') {
				text_cb(b);
				state = END;
			} else {
				char_cb(c);
			}
			break;
		case TAG:
			if (c == '>') {
				tag_cb(b);
				clearb();
				state = START;
			} else if (c == '\0') {
				error = 1;
				state = END;
			} else {
				char_cb(c);
			}
			break;
		case CHAR:
			if (c == ';') {
				t[m] = '\0';
				char_cb(extchar(t));
				state = START;
			} else if (c == '\0') {
				error = 1;
				state = END;
			} else {
				t[m++] = c;
			}
			break;
		default:
			error = 1;
			state = END;
		}
	}
	if (error) {
		fprintf(stderr, "Error\n");
	}
	return h;
}

